-------------------------------------------------------------------------------
-- ElvUI Chat Tweaks By Lockslap (US, Bleeding Hollow)
-- <Borderline Amazing>, http://ba-guild.com
-- Based on functionality provided by Prat and/or Chatter
-------------------------------------------------------------------------------
local E, C, L, DB		= unpack(ElvUI)
local _G				= getfenv(0)
local AceConfigDialog	= LibStub("AceConfigDialog-3.0")
local LSM				= LibStub("LibSharedMedia-3.0")
if IsAddOnLoaded("Prat") or IsAddOnLoaded("Chatter") or not C["chat"].enable then return end

-- load globals
ElvUI_ChatTweaks				= LibStub("AceAddon-3.0"):NewAddon("ElvUI_ChatTweaks", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")
ElvUI_ChatTweaks.version		= GetAddOnMetadata("ElvUI_ChatTweaks", "Version")
ElvUI_ChatTweaks.addon			= "|cff1784d1ElvUI Chat Tweaks|r version |cff00ff00" .. ElvUI_ChatTweaks.version .. "|r"

-- lua api
local format	= string.format
local sub		= string.sub
local gsub		= string.gsub
local match		= string.match
local find		= string.find
local insert	= table.insert
local concat	= table.concat
local sort		= table.sort

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
-- misc
local soundDesc				= "Sound to play when a message in %s is received.\n\n|cff00ff00To disable set to \"None\"|r."
local debug					= false
local clearCMD				= "|cff00ff00%s|r or |cff00ff00%s|r %s"
local db

-- hyperlinks
local hyperlinkTypes		= {}

-- channel names
local channels				= {}
local functions				= {}
local empty_tag				= "$$EMPTY$$"

-- editbox history
local history				= {}
local EditBox				= DEFAULT_CHAT_FRAME.editBox
local editEnabled			= false

-- name coloring
local cache					= {}
local local_names			= {} -- if not saving
local player				= UnitName("player")
local classes				= {					
	DRUID		= "Druid",
	MAGE		= "Mage",
	PALADIN		= "Paladin",
	PRIEST		= "Priest",
	ROGUE		= "Rogue",
	HUNTER		= "Hunter",
	SHAMAN		= "Shaman",
	WARLOCK		= "Warlock",
	WARRIOR		= "Warrior",
	DEATHKNIGHT	= "Death Knight",
}

-- chat sounds
local soundChannelString	= "CHAT_MSG_%s"
local soundChannelPattern	= "CHAT_MSG_(.+)"
local soundChannels 		= {
	["Guild"]				= "GUILD",
	["Officer"] 			= "OFFICER",
	["Party"] 				= "PARTY",
	["Party Leader"] 		= "PARTY_LEADER",
	["Raid"] 				= "RAID",
	["Raid Leader"] 		= "RAID_LEADER",
	["Say"] 				= "SAY",
	["Yell"]				= "YELL",
	["Battleground"]		= "BATTLEGROUND",
	["Battleground Leader"]	= "BATTLEGROUND_LEADER",
	["Emote"] 				= "EMOTE",
	["Monster Emote"] 		= "MONSTER_EMOTE",
	["Monster Say"]			= "MONSTER_SAY",
	["Raid Boss Emote"] 	= "RAID_BOSS_EMOTE"
}

-- chat filtering
local complaintAdded 	= COMPLAINT_ADDED
local prevReportTime	= 0
local prevLineID		= 0
local prevMessage		= nil
local prevPlayer		= nil
local filterResult		= nil

-- scroll reminder
local buttonsEnabled	= false

-- roll frame
local rollLastMSG		= nil
local rollPattern		= "(%a+) rolls (%d+) %((%d+)%-(%d+)%)"
local rollFormat		= "|cff%s%s rolls %d (%d-%d)|r"

-------------------------------------------------------------------------------
-- Hyperlink scripts
-------------------------------------------------------------------------------
local function buildHyperlinks()
	hyperlinkTypes = {
		item			= db.item,
		spell			= db.spell,
		unit			= db.unit,
		quest			= db.quest,
		enchant			= db.enchant,
		achievement		= db.achievement,
		instancelock	= db.instancelock,
		talent			= db.talent,
		glyph			= db.glyph,
	}
end

function ElvUI_ChatTweaks:OnHyperlinkEnter(self, refString)
	local linkToken = refString:match("^([^:]+)")
	if db.combat and InCombatLockdown() then return end
	if hyperlinkTypes[linkToken] then
		ShowUIPanel(GameTooltip)
		GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
		GameTooltip:SetHyperlink(refString)
		GameTooltip:Show()
	end
end

function ElvUI_ChatTweaks:OnHyperlinkLeave(self, refString)
	local linkToken = refString:match("^([^:]+)")
	if hyperlinkTypes[linkToken] then
		HideUIPanel(GameTooltip)
	end
end

-------------------------------------------------------------------------------
-- Channel name functions
-------------------------------------------------------------------------------
local function replaceChannel(origChannel, msg, num, channel)
	local f = functions[channel] or functions[channel:lower()]
	local newChannelName = f and f(channel) or channels[channel] or channels[channel:lower()] or msg
	if newChannelName == empty_tag then return "" end
	return ("|Hchannel:%s|h%s|h%s"):format(origChannel, newChannelName, " ")
end

local function replaceChannelRW(msg, channel)
	local f = functions[channel] or functions[channel:lower()]
	local newChannelName = f and f(channel) or channels[channel] or channels[channel:lower()] or msg
	return newChannelName .. " "
end

local function buildChannels()
	channels = {
		["Guild"]				= db.guild and db.guildText or "[Guild]",
		["Officer"]				= db.officer and db.officerText or "[Officer]",
		["Party"]				= db.party and db.partyText or "[Party]",
		["Party Leader"]		= db.partyLeader and db.partyLeaderText or "[Party Leader]",
		["Dungeon Guide"]		= db.dungeonGuide and db.dungeonGuideText or "[Dungeon Guide]",
		["Raid"]				= db.raid and db.raidText or "[Raid]",
		["Raid Leader"]			= db.raidLeader and db.raidLeaderText or "[Raid Leader]",
		["Raid Warning"]		= db.raidWarning and db.raidWarningText or "[Raid Warning]",
		["Battleground"]		= db.battleground and db.battlegroundText or "[Battleground]",
		["Battleground Leader"]	= db.battlegroundLeader and db.bgLeaderText or "[Battleground Leader]",	
	}
	functions = {}
	-- for channel name replacement
	for k, v in pairs(channels) do
		if v:match("^function%(") then
			functions[k] = loadstring("return " .. v)()
		end
	end
end

-------------------------------------------------------------------------------
-- Group Say slash command
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:SendChatMessage(text)
	SendChatMessage(text, self:GetGroupInformation())
end

function ElvUI_ChatTweaks:GetGroupInformation(slash)
	local inInstance, kind = IsInInstance()
	if inInstance and kind == "pvp" then
		return slash and "/bg" or "BATTLEGROUND"
	end
	if GetNumRaidMembers() > 0 then
		return slash and "/ra" or "RAID"
	elseif GetNumPartyMembers() > 0 then
		return slash and "/p" or "PARTY"
	end
	return slash and "/s" or "SAY"
end

-------------------------------------------------------------------------------
-- Invite handlers
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:SetItemRef(link, text, button)
	if db.altInvite and IsAltKeyDown() and sub(link, 1, 6) == "player" then
		local name = match(link, "player:([^:]+)")
		if name == player then return nil end -- dont need someone trying to invite themselves
		InviteUnit(name)
		return nil
	end
	return self.hooks.SetItemRef(link, text, button)
end

-------------------------------------------------------------------------------
-- Editbox history
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:AddHistoryLine(object, line)
	if not db.editboxHistory or not editEnabled then return end
	tinsert(history, line)
	-- clean things up
	for i = 1, #history - object:GetHistoryLines() do
		tremove(history, 1)
	end
end

local function LoadHistory()
	for _, line in ipairs(history) do
		EditBox:AddHistoryLine(line)
	end
end

-------------------------------------------------------------------------------
-- Justification
-------------------------------------------------------------------------------
local function SetJustification(reset)
	for i = 1, NUM_CHAT_WINDOWS do
		local ChatFrame = _G["ChatFrame" .. i]
		local justify = "LEFT"
		if not reset and db["FRAME_" .. i] then
			justify = db["FRAME_" .. i]
		end
		ChatFrame:SetJustifyH(justify)
	end
end

-------------------------------------------------------------------------------
-- Name coloring
-------------------------------------------------------------------------------
local getNameColor
do
	local sq2 = sqrt(2)
	local pi = _G.math.pi
	local cos = _G.math.cos
	local fmod = _G.math.fmod
	local strbyte = _G.strbyte
	local t = {}

	-- http://www.tecgraf.puc-rio.br/~mgattass/color/HSVtoRGB.htm


	local function HSVtoRGB(h, s, v)
	   if ( s == 0 ) then
		  --achromatic=fail
		  t.r = v
		  t.g = v
		  t.b = v
		  if not t.r then t.r = 0 end
		  if not t.g then t.g = 0 end
		  if not t.b then t.b = 0 end
		  return t.r,t.g,t.b
	   end
	   h = h/60
	   local i = floor(h)
	   local i1 = v * (1 - s)
	   local i2 = v * (1 - s * (h - i))
	   local i3 = v * (1 - s * (1 - (h - i)))
	   if i == 0 then
		  --	return v, i3, i1
		  t.r = v
		  t.g = i3
		  t.b = i1
	   elseif i == 1 then
		  --	return i2, v, i1
		  t.r = i2
		  t.g = v
		  t.b = i1
	   elseif i == 2 then
		  --	return i1, v, i3
		  t.r = i1
		  t.g = v
		  t.b = i3
	   elseif i == 3 then
		  --	return i3, i2, v
		  t.r = i3
		  t.g = i2
		  t.b = v
	   elseif i == 4 then
		  --	return i3, i1, v
		  t.r = i3
		  t.g = i1
		  t.b = v
	   elseif i == 5 then
		  --	return v, i1, i2
		  t.r = v
		  t.g = i1
		  t.b = i2
	   else
		  DEFAULT_CHAT_FRAME:AddMessage("ElvUI_ChatTweaks HSVtoRGB failed")
	   end
	   if not t.r then t.r = 0 end
	   if not t.g then t.g = 0 end
	   if not t.b then t.b = 0 end
	   return t.r,t.g,t.b
	end

	function getNameColor(name)
		local seed = 5381 --old seed: 5124
		local h, s, v = 1, 1, 1
		local r, g, b
		for i = 1, #name do
			seed = 33 * seed + strbyte(name, i) --used to use 29 here
		end
		-- h = fmod(seed, 255) / 255
		h = fmod(seed, 360) --changed the HSVtoRGB to acompany this change
		if (h > 220) and (h < 270) then
		   h = h + 60
		end
		t.r, t.g, t.b = HSVtoRGB(h, s, v)

		return t
	end
end

local function wipeCache()
	wipe(cache)
end

local function updateSaveData(val)
	if val then
		for k, v in pairs(local_names) do
			self.self.db.realm.names[k] = v
		end
	end
end

function ElvUI_ChatTweaks:AddPlayer(name, class, level, save)
	if name and class and class ~= UNKNOWN then
		if save or self.db.realm.names[name] then
			self.db.realm.names[name] = self.db.realm.names[name] or {}
			self.db.realm.names[name].class = class
			if level and level ~= 0 then
				self.db.realm.names[name].level = level
			end
		else
			local_names[name] = local_names[name] or {}
			local_names[name].class = class
			if level and level ~= 0 then
				local_names[name].level = level
			end
		end
	end
end

function ElvUI_ChatTweaks:FRIENDLIST_UPDATE(event)
	for i = 1, GetNumFriends() do
		local name, level, class = GetFriendInfo(i)
		self:AddPlayer(name, classes[class], level, db.saveFriend)
	end
end

function ElvUI_ChatTweaks:GUILD_ROSTER_UPDATE(event)
	if not IsInGuild() then return end
	for i = 1, GetNumGuildMembers() do
		local name, _, _, level, _, _, _, _, _, _, class = GetGuildRosterInfo(i)
		self:AddPlayer(name, class, level, db.saveGuild)
	end
end

function ElvUI_ChatTweaks:RAID_ROSTER_UPDATE(event)
	if not GetNumRaidMembers() or GetNumRaidMembers() == 0 then return end
	for i = 1, GetNumRaidMembers() do
		local name, _, _, level, _, class = GetRaidRosterInfo(i)
		self:AddPlayer(name, class, level, db.saveGroup)
	end
end

function ElvUI_ChatTweaks:PARTY_MEMBERS_CHANGED(event)
	if not GetNumPartyMembers() or GetNumPartyMembers() == 0 then return end
	for i = 1, GetNumPartyMembers() do
		local name, level	= UnitName("party" .. i), UnitLevel("party" .. i)
		local _, class		= UnitClass("party" .. i)
		self:AddPlayer(name, class, level, db.saveGroup)
	end
end

function ElvUI_ChatTweaks:PLAYER_TARGET_CHANGED(event)
	if not UnitExists("target") or not UnitIsPlayer("target") or not UnitIsFriend("player", "target") then return end
	local name, level	= UnitName("target"), UnitLevel("target")
	local _, class		= UnitClass("target")
	self:AddPlayer(name, class, level, db.saveTarget)
end

function ElvUI_ChatTweaks:UPDATE_MOUSEOVER_UNIT(event)
	if not UnitExists("mouseover") or not UnitIsPlayer("mouseover") or not UnitIsFriend("player", "mouseover") then return end
	local name, level	= UnitName("mouseover"), UnitLevel("mouseover")
	local _, class		= UnitClass("mouseover")
	self:AddPlayer(name, class, level, db.saveTarget)
end

function ElvUI_ChatTweaks:WHO_LIST_UPDATE(event)
	local whoResults = GetNumWhoResults()
	if whoResults > 0 and (whoResults <= 3 or db.saveAllWho) then
		for i = 1, whoResults do
			local name, _, level, _, _, _, class = GetWhoInfo(i)
			self:AddPlayer(name, class, level, db.saveWho)
		end
	end
end

function ElvUI_ChatTweaks:GetColor(className, isLocal)
	if isLocal then
		local found
		for k,v in pairs(LOCALIZED_CLASS_NAMES_FEMALE) do
			if v == className then className = k found = true break end
		end
		if not found then
			for k,v in pairs(LOCALIZED_CLASS_NAMES_MALE) do
				if v == className then className = k break end
			end
		end
	end
	local tbl = CUSTOM_CLASS_COLORS and CUSTOM_CLASS_COLORS[className] or RAID_CLASS_COLORS[className]
	local color = ("%02x%02x%02x"):format(tbl.r * 255, tbl.g * 255, tbl.b * 255)
	return color
end

local function changeName(msgHeader, name, extra, msgCount, displayName, msgBody)
	local names = ElvUI_ChatTweaks.db.realm.names
	if name == player then
		if db.nameColorSelf then
			msgBody = msgBody:gsub("("..player..")" , "|cffffff00>|r%1|cffffff00<|r"):gsub("("..player:lower()..")" , "|cffffff00>|r%1|cffffff00<|r")
		end
		if db.nameEmphasizeSelf then
			msgBody = msgBody:gsub("("..player..")" , "|cffff0000%1|r"):gsub("("..player:lower()..")" , "|cffff0000%1|r")
		end	
	end
	
	if not strmatch(displayName, "|cff") then
		displayName = ElvUI_ChatTweaks:ColorName(name)
	end
	
	cache[name] = displayName
	
	local result = names[name] or local_names[name]
	local level
	if result then
		level = db.level and result.level or false
	end
	
	if level and (not db.levelMax or level ~= MAX_PLAYER_LEVEL) then
		if db.levelColor then
			local color = GetQuestDifficultyColor(level)
			level = ("|cff%02x%02x%02x%s|r"):format(color.r * 255, color.g * 255, color.b * 255, level)
		else
			if strmatch(displayName, "|cff......") then
				level = gsub(displayName, "((|cff......).-|r)", function (string, color)
					return ("%s%s|r"):format( color, level )
				end )
			end
		end
		displayName = ("%s%s%s"):format(db.levelLocation == "LEFT" and level or displayName, db.levelSeparator, db.levelLocation == "RIGHT" and level or displayName)
	end
	
	return ("|Hplayer:%s%s%s|h%s%s%s|h%s"):format(name, extra, msgCount, db.nameOpen, displayName, db.nameClose, msgBody)
end

function ElvUI_ChatTweaks:ColorName(name)
	local names = ElvUI_ChatTweaks.db.realm.names
	local result = names[name] or local_names[name]
	local class = result and result.class or nil
	
	if cache[name] then
		name = cache[name]
	else
		if db.nameColorBy ~= "NONE" then
			local color = db.nameDefaultColor
			if db.nameColorBy == "CLASS" then
				color = CUSTOM_CLASS_COLORS and CUSTOM_CLASS_COLORS[class] or RAID_CLASS_COLORS[class] or db.nameDefaultColor
			elseif db.nameColorBy == "NAME" then
				color = getNameColor(name)
			end
			
			name = ("|cff%02x%02x%02x%s|r"):format(color.r * 255, color.g * 255, color.b * 255, name)
		end
	end
	
	return name
end

-------------------------------------------------------------------------------
-- Channel coloring
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:AddChannels(...)
	for i = 1, select("#", ...), 2 do
		local id, name = select(i, ...)
		db.channelColors[name] = db.channelColors[name] or {}
		if not db.channelColors[name].r then
			local r, g, b = GetMessageTypeColor(type(id) == "number" and ("CHANNEL" .. id) or id)
			db.channelColors[name].r 	= r
			db.channelColors[name].g 	= g
			db.channelColors[name].b 	= b
			db.channelColors[name].id	= id
		end
		if not self.options.args.general.args.channelColorOptions.args[name:gsub(" ", "_")] then
			self.options.args.general.args.channelColorOptions.args[name:gsub(" ", "_")] = {
				type	= "color",
				name	= name,
				desc	= "Select a color for this channel.",
				order	= type(id) == "number" and (50 + id) or 48,
				get		= function()
					local t = db.channelColors[name]
					if t then
						return t.r, t.g, t.b
					else
						return GetMessageTypeColor(type(id) == "number" and ("CHANNEL" .. id) or id)
					end
				end,
				set		= function(_, r, g, b)
					db.channelColors[name]		= db.channelColors[name] or {}
					db.channelColors[name].r 	= r
					db.channelColors[name].g	= g
					db.channelColors[name].b	= b
					db.channelColors[name].id	= id
					ChangeChatColor(type(id) == "number" and ("CHANNEL" .. id) or id, r, g, b)
				end
			}
		end
	end
end

function ElvUI_ChatTweaks:CHAT_MSG_CHANNEL_NOTICE(event, notice, _, _, fullname, _, _, channelType, channelNumber, channelName)
	if notice == "YOU_JOINED" then
		self:AddChannels(GetChannelList())
		channelName = channelName:match("^(%w+)")
		local c = db.channelColors[channelName]
		if c then
			ChangeChatColor("CHANNEL" .. channelNumber, c.r, c.g, c.b)
		end
	end
end

function ElvUI_ChatTweaks:UPDATE_CHAT_COLOR(event, channel, r, g, b)
	if channel then
		local num = tonumber(channel:match("(%d+)$"))
		local chanNumber = num and select(2, GetChannelName(num))
		local name = channelNum and chanNmber:match("^(%w+)") or channel
		db.channelColors[name]		= db.channelColors[name] or {}
		db.channelColors[name].r	= r
		db.channelColors[name].g	= g
		db.channelColors[name].b	= b
	end
end

-------------------------------------------------------------------------------
-- Channel sounds
-------------------------------------------------------------------------------
local function capitalize(str)
    return (str:gsub("^%l", string.upper))
end

local function getSettingName(str)
	if not str:find("_") then
		return str:lower() .. "Sound"
	else
		local first, temp = false, ""
		for word in str:gmatch("%w+") do
			if not first then
				temp = word:lower()
				first = true
			else
				temp = temp .. capitalize(word:lower())
			end
		end
		return temp .. "Sound"
	end
end

function ElvUI_ChatTweaks:ChannelSounds(event, ...)
	local channel = event:match(soundChannelPattern)
	if not db.channelSounds or not channel then return end
	local setting = getSettingName(channel)
	if db[setting] ~= "None" then
		PlaySoundFile(LSM:Fetch("sound", db[setting]))
	end
end

-------------------------------------------------------------------------------
-- Chat font settings
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:LSM_Registered(mediaType, name)
	if mediaType == LSM.MediaType.FONT then
		if name == db.fontFace then
			self:SetupChatFonts()
		end
	end
end

function ElvUI_ChatTweaks:RestoreChatFonts()
	for i = 1, NUM_CHAT_WINDOWS do
		local oldFont, CF = db.oldFonts["FRAME_" .. i], _G["ChatFrame" .. i]
		CF:SetFont(oldFont["face"], oldFont["size"], oldFont["outline"])
	end
end

function ElvUI_ChatTweaks:SetupChatFonts(face, size, outline)
	for i = 1, NUM_CHAT_WINDOWS do
		chatFrame = _G["ChatFrame" .. i]
		chatFrame:SetFont(LSM:Fetch("font", face), size, outline)
	end
end

-------------------------------------------------------------------------------
-- Scroll reminder
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:CreateBottomButton(frame)
	local button	= CreateFrame("Button", nil, frame)
	button:SetHeight(E.Scale(22))
	button:SetWidth(E.Scale(20))
	button:SetAlpha(0.75)
	button:SetPoint(frame:GetLeft() > (GetScreenWidth() / 2) and "BOTTOMLEFT" or "BOTTOMRIGHT", 0, 0)
	button:SetTemplate("Default", true)
	button:Hide()
	
	local buttontex = button:CreateTexture(nil, 'OVERLAY')
	buttontex:SetPoint('TOPLEFT', button, 'TOPLEFT', 2, -2)
	buttontex:SetPoint('BOTTOMRIGHT', button, 'BOTTOMRIGHT', -2, 2)
	buttontex:SetTexture([[Interface\AddOns\ElvUI\media\textures\arrowdown.tga]])
	
	button:SetScript("OnClick", function(self) self:GetParent():ScrollToBottom() end)
	button:SetScript("OnEnter", function() button:SetAlpha(1) end)
	button:SetScript("OnLeave", function() button:SetAlpha(0.75) end)
	
	frame.downButton = button
end

function ElvUI_ChatTweaks:EnableBottomButton()
	if not db.scrollReminder or buttonsEnabled then return end
	buttonsEnabled = true
	for i = 1, NUM_CHAT_WINDOWS do
		local frame = _G["ChatFrame" .. i]
		if frame then self:ApplyBottomButton(frame) end
	end
end

function ElvUI_ChatTweaks:DisableBottomButton()
	if buttonsEnabled then return end
	buttonsEnabled = false
	for i = 1, NUM_CHAT_WINDOWS do
		local frame = _G["ChatFrame" .. i]
		if frame then self:RemoveBottomButton(frame) end
	end
end

function ElvUI_ChatTweaks:ApplyBottomButton(frame)
	if self:IsHooked(frame, "ScrollUp") then return end
	self:Hook(frame, "ScrollUp", true)
	self:Hook(frame, "ScrollToTop", "ScrollUp", true)
	self:Hook(frame, "PageUp", "ScrollUp", true)
	self:Hook(frame, "ScrollDown", true)
	self:Hook(frame, "ScrollToBottom", "ScrollDownForce", true)
	self:Hook(frame, "PageDown", "ScrollDown", true)
	if frame:GetCurrentScroll() ~= 0 then frame.downButton:Show() end
end

function ElvUI_ChatTweaks:RemoveBottomButton(frame)
	self:Unhook(frame, "ScrollUp")
	self:Unhook(frame, "ScrollToTop")
	self:Unhook(frame, "PageUp")
	self:Unhook(frame, "ScrollDown")
	self:Unhook(frame, "ScrollToBottom")
	self:Unhook(frame, "PageDown")
	frame.downButton:Hide()
end

-- hook goodness
function ElvUI_ChatTweaks:ScrollUp(frame)
	frame.downButton:Show()
	frame.downButton:UnlockHighlight()
end

function ElvUI_ChatTweaks:ScrollDown(frame)
	if frame:GetCurrentScroll() == 0 then
		frame.downButton:Hide()
		frame.downButton:UnlockHighlight()
	end
end

function ElvUI_ChatTweaks:ScrollDownForce(frame)
	frame.downButton:Hide()
	frame.downButton:UnlockHighlight()
end

-------------------------------------------------------------------------------
-- Roll frame
-------------------------------------------------------------------------------
local FCF_FlashTab = function(self)
	local tab = _G[self:GetName() .. "TabFlash"]
	
	if self == SELECTED_CHAT_FRAME or UIFrameIsFlashing(tab) then
		return
	else
		UIFrameFlash(tab, 0.25, 0.25, -1, nil, 0.5, 0.5)
	end
end
local function RollFrameFilter(message)
	local roller, roll, minRoll, maxRoll = message:match(rollPattern)
	local frame = _G[db.rollRouteTo]
	local r, g, b = GetMessageTypeColor("CHAT_MSG_SYSTEM")
	if not frame then return end
	FCF_FlashTab(frame)
	if not db.rollColors then
		frame:AddMessage(message, r, g, b)
		return true
	else
		if roller:lower() == UnitName("player"):lower() then
			-- self roll
			if roll == "1" then
				-- you lose
				r = db.rollColorOwnLoss and db.rollOwnLossColor.r or db.rollColorOwn and db.rollOwnColor.r or r
				g = db.rollColorOwnLoss and db.rollOwnLossColor.g or db.rollColorOwn and db.rollOwnColor.g or g
				b = db.rollColorOwnLoss and db.rollOwnLossColor.b or db.rollColorOwn and db.rollOwnColor.b or b
			else
				r = db.rollColorOwn and db.rollOwnColor.r or db.rollColorReg and db.rollRegColor.r or r
				g = db.rollColorOwn and db.rollOwnColor.g or db.rollColorReg and db.rollRegColor.g or g
				b = db.rollColorOwn and db.rollOwnColor.b or db.rollColorRed and db.rollRegColor.b or b
			end
		else
			-- someone else's roll
			if roll == "1" then
				r = db.rollColorLoss and db.rollLossColor.r or db.rollColorReg and db.rollRegColor.r or r
				g = db.rollColorLoss and db.rollLossColor.g or db.rollColorReg and db.rollRegColor.g or g
				b = db.rollColorLoss and db.rollLossColor.b or db.rollColorReg and db.rollRegColor.b or b
			else
				r = db.rollColorReg and db.rollRegColor.r or r
				g = db.rollColorReg and db.rollRegColor.g or g
				b = db.rollColorReg and db.rollRegColor.b or b
			end
		end
		frame:AddMessage(message, r, g, b)
		return true
	end
end

-------------------------------------------------------------------------------
-- Spam filtering
-------------------------------------------------------------------------------
-- from BadBoy
local function IsSpam(message)
	local points, phish, strict, icon = 0, 0, false, false
	local trigs = ElvUI_ChatTweaks.triggers["GoldSelling"]
	for i = 1, #trigs do
		if find(message, trigs[i]) then
			if i > 89 then -- instant report
				points = points + 9
			elseif i > 65 and i < 90 then
				phish = phish + 1
			elseif i > 57 and i < 66 and not icon then
				points = points + 1
				icon = true
			elseif i > 51 and i < 58 and not strict then
				points = points + 2
				phish = phish + 1
				strict = true
			elseif i > 41 and i < 52 then
				points = points + 2
			elseif i > 7 and i < 42 then
				points = points + 1
			elseif i < 8 then
				-- remove points for safe words
				points = points - 2
				phish = phish - 2
			end
			if debug then print(i, trigs[i], message, points, phish) end
			if points > 3 or phish > 3 then
				return true
			end
		end
	end
end

local function FilterChat(_, channel, message, player, _, _, _, flag, channelID, _, _, _, lineID)
	if channel == "CHAT_MSG_AFK" and db.spamAFK then return true end
	if channel == "CHAT_MSG_DND" and db.spamDND then return true end
	if not db.spamFilter then return end
	local guildTrigs, iconTrigs = ElvUI_ChatTweaks.triggers["Guild"], ElvUI_ChatTweaks.triggers["RaidIcons"]
	if lineID == prevLineID then
		-- prevent duplicate messages from being reported
		return filterResult
	else
		prevLineID = lineID -- for above
		
		-- only scan officials channels (general, trade, etc.) not custom channels (minus guild recruitment)
		if channel == "CHAT_MSG_CHANNEL" and (channelID == 0 or channelID == 25) then filterResult = nil; return end
		-- dont scan ourselves/friends/GMs/guildies/etc.
		if not CanComplainChat(lineID) or UnitIsInMyGuild(player) or UnitInRaid(player) or UnitInParty(player) then filterResult = nil; return end
	
		-- check whispers
		if channel == "CHAT_MSG_WHISPER" then
			if flag == "GM" then filterResult = nil; return end -- dont scan GM messages
			
			for i = 1, select(2, BNGetNumFriends()) do
				local toon = BNGetNumFriendToons(i)
				for x = 1, toon do
					local _, name, game = BNGetFriendToonInfo(i, x)
					if name == player and game == "WoW" then filterResult = nil; return end
				end
			end
		end
	end
	
	-- powerleveling, gold selling, etc.
	if db.spamGoldSelling and IsSpam(message:lower():gsub(" ", "")) then
		
		-- dont report the same person twice
		if message == prevMessage and player == prevPlayer then filterResult = true; return true; end
		
		-- save for later
		prevMessage	= message
		prevPlayer	= player	
		
		local now = GetTime()
		-- dont report too fast!
		if (now - prevReportTime) > 0.5 then
			prevReportTime = now
			COMPLAINT_ADDED = "|cff33ff99ElvUI_ChatTweaks:|r " .. complaintAdded .. " |Hplayer:"..player.."|h["..player.."]|h"			
			if debug then 
				print(message, "-", channel, "-", player)
			else
				if db.spamReportConfirm then
					StaticPopupDialogs["CONFIRM_REPORT_SPAM_CHAT"].text = REPORT_SPAM_CONFIRMATION .. "\n\n" .. message:gsub("%%", "%%%%")
					local dialog = StaticPopup_Show("CONFIRM_REPORT_SPAM_CHAT", player)
					dialog.data = lineID
				else
					ComplainChat(lineID)
				end
			end
		end
		filterResult = true
		return true
		
	-- guild recruitment spam
	elseif db.spamGuildRecruit and channel == "CHAT_MSG_CHANNEL" then
		message = message:lower()
		for i = 1, #guildTrigs do
			if message:find(guildTrigs[i]) then
				if debug then print(message, "-", guildTrigs[i], "-", player) end
				filterResult = true
				return true
			end
		end
	
	-- keyword filtering
	elseif db.spamKeywords and #db.spamFilters > 0 and (channel == "CHAT_MSG_CHANNEL" or channel == "CHANNEL_MSG_YELL" or channel == "CHANNEL_MSG_SAY") then
		message = message:lower():gsub(" ", "")
		for i = 1, #db.spamFilters do
			local check = db.spamFilters[i]:lower():gsub(" ", "")
			if message:find(check) then
				if debug then print(message, "-", db.spamFilters[i], "-", player) end
				filterResult = true
				return true
			end
		end
	
	-- remove raid icons from channels they dont belong
	elseif db.spamRemoveIcons then
		for i = 1, #iconTrigs do
			message, found = message:gsub(iconTrigs[i], "")
			if found > 0 then modify = true end
		end
		-- only filter if the message was modified
		if modify then return false, message, player, _, _, _, flag, channelID, _, _, _, lineID end
	end
	
	filterResult = nil
end

-- system messages get their own filter
ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", function(_, _, message)
	if db.rollFrame and message:match(rollPattern) and message ~= rollLastMSG then rollLastMSG = message; return RollFrameFilter(message) end
	if not db.spamFilter or not db.spamGoldSelling then return false end
	if message == complaintAdded then
		return -- manual report, so dont do anything
	elseif message == COMPLAINT_ADDED then
		COMPLAINT_ADDED = complaintAdded
		
		if db.spamReportConfirm then
			StaticPopupDialogs["CONFIRM_REPORT_SPAM_CHAT"].text = REPORT_SPAM_CONFIRMATION
		end
		
		if db.spamReportSilent then
			return true
		end
	else
		SetCVar("spamFilter", 1) -- let blizzard's spam filter help us :-)
	end
end)

function ElvUI_ChatTweaks:PopulateFilters(filters)
	db.spamFilters = {}
	for index, value in pairs(filters) do
		if value ~= "" then -- just in case
			db.spamFilters[#db.spamFilters + 1] = value
		end
	end
end

function ElvUI_ChatTweaks:FiltersToString()
	local trigs = ""
	for i = 1, #db.spamFilters do
		if db.spamFilters[i]:trim() ~= "" then
			trigs = trigs .. db.spamFilters[i] .. "\n"
		end
	end
	return trigs
end

-------------------------------------------------------------------------------
-- Clear chat commands
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:ClearChat(frame)
	local frameType = type(frame)
	local type = frame:GetObjectType() or nil
	
	if type == "ScrollingMessageFrame" then
		frame:Clear()
	end
end

function ElvUI_ChatTweaks:ClearAllChats()
	for i = 1, NUM_CHAT_WINDOWS do
		self:ClearChat(_G["ChatFrame" .. i])
	end
end

-------------------------------------------------------------------------------
-- Global hooks
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:AddMessage(frame, text, ...)
	local id = frame:GetID()
	if not id or not text or type(text) ~= "string" then
		return self.hooks[frame].AddMessage(frame, text, ...)
	end
	
	-- timestamp replacements
	if db.timestamps then
		local color = ("%02x%02x%02x"):format(db.tsColor.r * 255, db.tsColor.g * 255, db.tsColor.b * 255)
		text = "|cff" .. color .. db.tsOpen .. date(db.tsFormat) .. db.tsClose .. "|r " .. text
	end
	
	-- channel name replacements
	if db.channels then
		text = text:gsub("|Hchannel:(%S-)|h(%[([%d. ]*)([^%]]+)%])|h ", replaceChannel)
		text = text:gsub("(%[(Raid Warning)%]) ", replaceChannelRW)
	end
	
	-- name coloring
	if db.names then
		text = text:gsub("(|Hplayer:([^|:]+)([:%d+]*)([^|]*)|h%[([^%]]+)%]|h)(.-)$", changeName)	
	end
	
	-- scroll reminder
	if db.scrollReminder then
		if frame:GetCurrentScroll() > 0 then
			frame.downButton:Show()
			frame.downButton:LockHighlight()
		else
			frame.downButton:Hide()
			frame.downButton:UnlockHighlight()
		end	
	end
	
	return self.hooks[frame].AddMessage(frame, text, ...)
end

function ElvUI_ChatTweaks:OnTextChanged(object)
	local text = object:GetText()
	if text:sub(1, 4) == "/gs" then
		object:SetText(self:GetGroupInformation(true) .. text:sub(5))
		ChatEdit_ParseText(object, 0)
	end
	self.hooks[object].OnTextChanged(object)
end

function ElvUI_ChatTweaks:GetPossibleRollFrames()
	local frames = {}
	for i = 1, NUM_CHAT_WINDOWS do
		local frame = _G["ChatFrame" .. i]
		if frame and frame ~= ChatFrame1 then
			frames["ChatFrame" .. i] = "ChatFrame" .. i 
		end
	end
	return frames
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------
function ElvUI_ChatTweaks:OnEnable()
	for i = 1, NUM_CHAT_WINDOWS do
		-- hyperlinks hovering
		local frame = _G["ChatFrame" .. i]
		if db.hyperlinks then
			self:RawHookScript(frame, "OnHyperlinkEnter")
			self:RawHookScript(frame, "OnHyperlinkLeave")
		end
		
		-- invites/timestamps/channel names
		if frame ~= COMBATLOG then
			self:RawHook(frame, "AddMessage", true)
		end
	end
	
	-- alt-click invite
	if db.altInvite then
		self:RawHook(nil, "SetItemRef", true)
	end
	
	-- groupsay chat command
	if db.groupSay then
		self:HookScript(EditBox, "OnTextChanged")
		if not self.slashCommandRegistered then -- make sure it doesnt register twice
			self:RegisterChatCommand("gs", "SendChatMessage")
			self.slashCommandRegistered = true
		end
	end
	
	-- chat logging
	self.isLogging = LoggingChat()
	if db.chatLogging and not self.isLogging then
		LoggingChat(true)
	end
	
	-- editbox history
	history = self.db.char.history
	LoadHistory()
	editEnabled = true
	self:SecureHook(EditBox, "AddHistoryLine")
	
	-- text justification
	if db.justify then
		SetJustification(false)
	end
	
	-- name coloring
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("WHO_LIST_UPDATE")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "WHO_LIST_UPDATE")
	self:RegisterEvent("FRIENDLIST_UPDATE")
	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	if db.names and IsInGuild() then
		GuildRoster()
	end
	
	-- channel coloring
	self:RegisterEvent("UPDATE_CHAT_COLOR")
	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE")
	self:AddChannels(GetChannelList())
	self:AddChannels(
		"SAY", "Say",
		"YELL", "Yell",
		"GUILD", "Guild", 
		"OFFICER", "Officer", 
		"PARTY", "Party", 
		"PARTY_LEADER", PARTY_LEADER,
		"RAID", "Raid", 
		"RAID_LEADER", "Raid Leader",
		"RAID_WARNING", "Raid Warning",
		"BATTLEGROUND", "Battleground",
		"BATTLEGROUND_LEADER", "Battleground Leader",
		"WHISPER", "Whisper",
		"BN_WHISPER", "RealID Whisper",
		"BN_CONVERSATION", "RealID Conversation"
	)
	
	-- scroll reminder
	if db.scrollReminder then self:EnableBottomButton() end
end

function ElvUI_ChatTweaks:OnDisable()
	self:UnregisterAllEvents()
	self:UnhookAll()
	
	if db.chatFonts then
		self:RestoreChatFonts()
	end
	
	if db.scrollReminder then
		self:DisableBottomButton()
	end
	
	for _, value in pairs(soundChannels) do
		self:UnregisterEvent(format(soundChannelString, value))
	end

	SetJustification(true)
end

function ElvUI_ChatTweaks:OnProfileChanged(event, database, newProfileKey)
	db = database.profile -- update the db to new settings
end

function ElvUI_ChatTweaks:OnInitialize()
	-- for the settings
	self.db = LibStub("AceDB-3.0"):New("ElvUI_ChatTweaksDB", self.defaults)
	db = self.db.profile
	
	-- load justification options into the options table
	for i = 1, NUM_CHAT_WINDOWS do
		local s, f = "FRAME_" .. i, _G["ChatFrame" .. i]
		
		-- save current chat font, then set the new stuff
		if db.chatFonts then
			local face, size, outline = f:GetFont()
			db.oldFonts[s] = {
				face	= face,
				size	= size,
				outline	= outline
			}
		end
		
		self.options.args.general.args.justifyOptions.args[s] = {
			type	= "select",
			name	= "Chat Frame " .. i,
			desc	= "Chat Frame " .. i,
			values	= {
				LEFT	= "Left",
				RIGHT	= "Right",
				CENTER	= "Center"
			},
			get		= function() return db[s] or "LEFT" end,
			set		= function(_, value)
				db[s] = value
				f:SetJustifyH(value)
			end
		}
	end
	
	--LSM.RegisterCallback(self, "LibSharedMedia_Registered", "LSM_Registered")
	--LSM.RegisterCallback(self, "LibSharedMedia_SetGlobal", "LSM_Registered")
	ElvUI_ChatTweaks:SetupChatFonts(db.fontFace, db.fontSize, db.fontOutline)
	
	-- channel sounds
	for name, value in pairs(soundChannels) do
		-- register event
		self:RegisterEvent(format(soundChannelString, value), "ChannelSounds")
		
		local settingName = getSettingName(value)
		if not self.options.args.general.args.channelSoundOptions.args[settingName] then
			self.options.args.general.args.channelSoundOptions.args[settingName] = {
				type			= "select",
				order			= 1,
				dialogControl	= "LSM30_Sound",
				name			= name,
				desc			= format(soundDesc, name:lower()),
				values			= AceGUIWidgetLSMlists.sound,
				get				= function() return db[settingName] or "None" end,
				set				= function(_, change) db[settingName] = change; end
			}
		end
	end

	self.options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ElvUI_ChatTweaks", self.options)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ElvUI_ChatTweaksSlashCommand", self.optionsSlash, {"chattweats", "ct"})
	self.frames = {}
	self.frames.general	= AceConfigDialog:AddToBlizOptions("ElvUI_ChatTweaks", "ElvUI_ChatTweaks", nil, "general")
	self.frames.profile	= AceConfigDialog:AddToBlizOptions("ElvUI_ChatTweaks", "Profiles", "ElvUI_ChatTweaks", "profiles")
	self.frames.about	= LibStub("LibAboutPanel").new("ElvUI_ChatTweaks", "ElvUI_ChatTweaks")
	
	-- profile handling
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	
	-- for channel name replacement
	buildChannels()
	
	-- hyperlink goodness
	buildHyperlinks()
	
	-- disable blizzard's ugly timestamps
	if db.timestamps and GetCVar("showTimestamps") ~= "none" then
		SetCVar("showTimestamps", "none")
	end
	
	-- chat filtering
	local filterChannels = {
		"CHAT_MSG_CHANNEL",
		"CHAT_MSG_SAY",
		"CHAT_MSG_YELL",
		"CHAT_MSG_WHISPER",
		"CHAT_MSG_EMOTE",
		"CHAT_MSG_DND",
		"CHAT_MSG_AFK"
	}
	for _, channel in pairs(filterChannels) do
		ChatFrame_AddMessageEventFilter(channel, FilterChat)
	end
	filterChannels = nil
	
	-- setup the scroll button
	for i = 1, NUM_CHAT_WINDOWS do
		local frame		= _G["ChatFrame".. i]
		self:CreateBottomButton(frame)
	end
	
	-- just because i think the command should be "/gkick"
	if CanGuildRemove() then
		self:RegisterChatCommand("gkick", function(args)
			local whoToKick = self:GetArgs(args)
			if not whoToKick then 
				print("|cffffff00Usage: /gkick <name>|r")
			elseif UnitIsInMyGuild(whoToKick) then 
				GuildUninvite(whoToKick)
			end
		end)
	end
	
	-- clear chat command
	if db.clearChat then
		self:RegisterChatCommand("clear", function() ElvUI_ChatTweaks:ClearChat(SELECTED_CHAT_FRAME) end)
		self:RegisterChatCommand("clr",	function() ElvUI_ChatTweaks:ClearChat(SELECTED_CHAT_FRAME) end)
		self:RegisterChatCommand("clearall", function() ElvUI_ChatTweaks:ClearAllChats() end)
		self:RegisterChatCommand("clrall", function() ElvUI_ChatTweaks:ClearAllChats() end)
	end
end

-------------------------------------------------------------------------------
-- Configuration options
-------------------------------------------------------------------------------
-- options table
ElvUI_ChatTweaks.options = {
	type = "group",
	name = "ElvUI Chat Tweaks",
	args = {
		general = {
			type		= "group",
			name		= "General Settings",
			--order		= 1,
			args		= {
				description	= {
					type	= "description",
					order	= 1,
					name	= ElvUI_ChatTweaks.addon .. " is designed to add a lot of the functionality of full fledged chat addons like Prat or Chatter, but without a lot of the unneeded bloat.  I wrote it to be as lightweight as possible, while still powerful enough to accomplish it's intended function.\n",
				},
				generalOptions = {
					type	= "group",
					name	= "General Options",
					order	= 2,
					args	= {
						hyperlinks = {
							type	= "toggle",
							--order	= 1,
							name	= "Hyperlink Hovering",
							desc	= "Toggle ability to show tooltip of a hyperlink by hovering over it in the chat frames.",
							get		= function() return db.hyperlinks end,
							set		= function(_, value)
								db.hyperlinks = value
								for i = 1, NUM_CHAT_WINDOWS do
									local frame = _G["ChatFrame" .. i]
									if not db.hyperlinks then
										if ElvUI_ChatTweaks:IsHooked(frame, "OnHyperlinkEnter") then ElvUI_ChatTweaks:Unhook(frame, "OnHyperlinkEnter") end
										if ElvUI_ChatTweaks:IsHooked(frame, "OnHyperlinkLeave") then ElvUI_ChatTweaks:Unhook(frame, "OnHyperlinkLeave") end
									else
										ElvUI_ChatTweaks:RawHookScript(frame, "OnHyperlinkEnter")
										ElvUI_ChatTweaks:RawHookScript(frame, "OnHyperlinkLeave")
									end
								end
							end,
						},
						timestamps = {
							type	= "toggle",
							--order	= 2,
							name	= "Show Timestamps",
							desc	= "Add a timestamp to each chat line.",
							get		= function() return db.timestamps end,
							set		= function(_, value) 
								db.timestamps = value
								
								if db.timestamps and GetCVar("showTimestamps") ~= "none" then
									SetCVar("showTimestamps", "none")
								end
							end,
						},
						channels = {
							type	= "toggle",
							--order	= 3,
							name	= "Shorten Channel Names",
							desc	= "Shorten the channel names that ElvUI doesn't do by default.  For example Guild -> G.",
							get		= function() return db.channels end,
							set		= function(_, value) db.channels = value end,
						},
						groupSay = {
							type	= "toggle",
							--order	= 4,
							name	= "Group Say Command",
							desc	= "Add a group say command to automatically to speak in your group (party, raid, battleground).",
							get		= function() return db.groupSay end,
							set		= function(_, value) 
								db.groupSay = value
								
								-- (un)hook the event
								if db.groupSay then
									ElvUI_ChatTweaks:HookScript(EditBox, "OnTextChanged")
								elseif not db.groupSay then
									ElvUI_ChatTweaks:Unhook(EditBox, "OnTextChanged")
								end
									
								-- (un)register the chat command
								if db.groupSay and not ElvUI_ChatTweaks.slashCommandRegistered then
									ElvUI_ChatTweaks:RegisterChatCommand("gs", "SendChatMessage")
									ElvUI_ChatTweaks.slashCommandRegistered = true
								elseif not db.groupSay and ElvUI_ChatTweaks.slashCommandRegistered then
									ElvUI_ChatTweaks:UnregisterChatCommand("gs")
									ElvUI_ChatTweaks.slashCommandRegistered = false
								end
							end,
						},
						chatLogging = {
							type		= "toggle",
							--order		= 5,
							name		= "Automatic Chat Logging",
							desc		= "Automatically enable chat logging.",
							get			= function() return db.chatLogging end,
							set			= function(_, value)
								db.chatLogging = value
								LoggingChat(value)
							end,
						},
						altInvite = {
							type		= "toggle",
							--order		= 6,
							name		= "Alt-Click Invite",
							desc		= "Alt-Click someone's name to invite them to your group.",
							get			= function() return db.altInvite end,
							set			= function(_, value) db.altInvite = value end,
						},
						editboxHistory = {
							type		= "toggle",
							--order		= 7,
							name		= "Editbox History",
							desc		= "Remembers the history of the editbox between sessions.",
							get			= function() return db.editboxHistory end,
							set			= function(_, value)
								db.editboxHistory = value
								if value then
									LoadHistory()
								end
							end
						},
						justify = {
							type		= "toggle",
							--order		= 8,
							name		= "Text Justification",
							desc		= "Toggle text justification in the chatframes.",
							get			= function() return db.justify end,
							set			= function(_, value)
								db.justify = value
								SetJustification(not value)
							end
						},
						names = {
							type		= "toggle",
							--order		= 9,
							name		= "Color Names",
							desc		= "Colors names based on your chosen method.",
							get			= function() return db.names end,
							set			= function(_, value) db.names = value end
						},
						channelSaveColors = {
							type		= "toggle",
							--order		= 10,
							name		= "Channel Colors",
							desc		= "Remember channel colors by name, rather than number.",
							get			= function() return db.channelSaveColors end,
							set			= function(_, value) db.channelSaveColors = value end,
						},
						channelSounds = {
							type		= "toggle",
							----order		= 12,
							name		= "Channel Sounds",
							desc		= "Play a sound when a message is received in a particular channel.",
							get			= function() return db.channelSounds end,
							set			= function(_, value) db.channelSounds = value end,
						},
						chatFonts = {
							type		= "toggle",
							name		= "Chat Fonts",
							desc		= "Set your own chat font, independant of ElvUI's settings.",
							get			= function() return db.chatFonts end,
							set			= function(_, value)
								db.chatFonts = value
								if not db.chatFonts then
									ElvUI_ChatTweaks:RestoreChatFonts()
								else
									ElvUI_ChatTweaks:SetupChatFonts(db.fontFace, db.fontSize, db.fontOutline)
								end
							end,
						},
						spamFilter		= {
							type		= "toggle",
							name		= "Spam Filter",
							desc		= "Filters various channels looking for specific keywords and phrases.",
							get			= function() return db.spamFilter end,
							set			= function(_, value) db.spamFilter = value end,
						},
						scrollReminder	= {
							type		= "toggle",
							name		= "Scroll Reminder",
							desc		= "Shows a button to remind you that you have scrolled up.",
							get			= function() return db.scrollReminder end,
							set			= function(_, value) 
								db.scrollReminder = value
								if value then
									ElvUI_ChatTweaks:EnableBottomButton()
								else
									ElvUI_ChatTweaks:DisableBottomButton()
								end
							end
						},
						rollFrame		= {
							type		= "toggle",
							name		= "Roll Frame",
							desc		= "Moves any rolls to a separate chat frame, and will color your frames differently.\n\nThis is more a feature for me, but you're all welcome to it.  I use this for playing the roll game for gold.",
							get			= function() return db.rollFrame end,
							set			= function(_, value) db.rollFrame = value end,
						},
						clearChat		= {
							type		= "toggle",
							name		= "Clear Chat Command",
							desc		= "Adds commands to clear the chat frames.\n\n" .. format(clearCMD, "/clear", "/clr", "Clear current chat.") .. "\n" .. format(clearCMD, "/clearall", "/clrall", "Clears all chats."),
							get			= function() return db.clearChat end,
							set			= function(_, value)
								db.clearChat = value
								if value then
									ElvUI_ChatTweaks:RegisterChatCommand("clear", function() ElvUI_ChatTweaks:ClearChat(SELECTED_CHAT_FRAME) end)
									ElvUI_ChatTweaks:RegisterChatCommand("clr",	function() ElvUI_ChatTweaks:ClearChat(SELECTED_CHAT_FRAME) end)
									ElvUI_ChatTweaks:RegisterChatCommand("clearall", function() ElvUI_ChatTweaks:ClearAllChats() end)
									ElvUI_ChatTweaks:RegisterChatCommand("clrall", function() ElvUI_ChatTweaks:ClearAllChats() end)
								else
									ElvUI_ChatTweaks:UnregisterChatCommand("clear")
									ElvUI_ChatTweaks:UnregisterChatCommand("clr")
									ElvUI_ChatTweaks:UnregisterChatCommand("clearall")
									ElvUI_ChatTweaks:UnregisterChatCommand("clrall")
								end
							end,
						},
					}
				},
				chatFontOptions = {
					type		= "group",
					name		= "Chat Font Options",
					disabled	= function() return not db.chatFonts end,
					args		= {
						fontFace	= {
							type			= "select",
							dialogControl	= "LSM30_Font",
							name			= "Font Face",
							desc			= "Set the default font face for the chat frames.",
							values			= AceGUIWidgetLSMlists.font,
							get				= function() return db.fontFace end,
							set				= function(_, value)
								db.fontFace = value
								ElvUI_ChatTweaks:SetupChatFonts(value, db.fontSize, db.fontOutline)
							end,
						},
						fontOutline	= {
							type			= "select",
							name			= "Font Outline",
							desc			= "Set the outline of the font.",
							values			= {[""] = "None", ["OUTLINE"] = "Outline", ["THICKOUTLINE"] = "Thick Outline"},
							get				= function() return db.fontOutline end,
							set				= function(_, value)
								db.fontOutline = value
								ElvUI_ChatTweaks:SetupChatFonts(db.fontFace, db.fontSize, value)
							end,
						},
						fontSize	= {
							type			= "range",
							name			= "Font Size",
							desc			= "Set the font size.",
							get				= function() return db.fontSize end,
							set				= function(_, value)
								db.fontSize = value
								ElvUI_ChatTweaks:SetupChatFonts(db.fontFace, value, db.fontOutline)
							end,
							min = 4, max = 30, step = 1, bigStep = 1
						},
						fontReset	= {
							type			= "execute",
							name			= "Reset All Fonts",
							desc			= "Resets all fonts to their original settings.",
							func			= function()
								for i = 1, NUM_CHAT_WINDOWS do
									local CF = _G["ChatFrame" .. i]
									local oldFont = db.oldFonts["FRAME_" .. i]
									CF:SetFont(oldFont["face"], oldFont["size"], oldFont["outline"])
								end
								db.fontFace		= db.oldFonts["FRAME_1"]["face"]
								db.fontSize		= db.oldFonts["FRAME_1"]["size"]
								db.fontOutline	= db.oldFonts["FRAME_1"]["outline"]
							end,
						},
					}
				},
				channelColorOptions = {
					type		= "group",
					----order		= 99,
					name		= "Channel Color Options",
					disabled	= function() return not db.channelSaveColors end,
					args		= {
						splitter = {	-- for separating channels
							type		= "header",
							name		= "Other Channels",
							order		= 49,
						}
					}
				},
				channelOptions = {
					type		= "group",
					----order		= 100,
					name		= "Channel Name Options",
					--guiInline	= true,
					disabled	= function() return not db.channels end,
					args		= {
						guild				= {
							type		= "toggle",
							order		= 1,
							name		= "Guild",
							desc		= "Guild Channel",
							get			= function() return db.guild end,
							set			= function(_, value) db.guild = value; buildChannels() end
						},
						guildText			= {
							type		= "input",
							order		= 2,
							name		= "Guild Text",
							desc		= "Text for guild chat.",
							disabled	= function() return not db.channels or not db.guild end,
							get			= function() return db.guildText end,
							set			= function(_, value) db.guildText = value; buildChannels() end,
						},
						officer				= {
							type		= "toggle",
							order		= 3,
							name		= "Officer",
							desc		= "Officer Channel",
							get			= function() return db.officer end,
							set			= function(_, value) db.officer = value; buildChannels() end
						},
						officerText			= {
							type		= "input",
							order		= 4,
							name		= "Officer Text",
							desc		= "Text for officer chat.",
							disabled	= function() return not db.channels or not db.officer end,
							get			= function() return db.officerText end,
							set			= function(_, value) db.officerText = value; buildChannels() end
						},
						party				= {
							type		= "toggle",
							order		= 5,
							name		= "Party",
							desc		= "Party Channel",
							get			= function() return db.party end,
							set			= function(_, value) db.party = value; buildChannels() end
						},
						partyText			= {
							type		= "input",
							order		= 6,
							name		= "Party Text",
							desc		= "Text for party chat.",
							disabled	= function() return not db.channels or not db.party end,
							get			= function() return db.partyText end,
							set			= function(_, value) db.partyText = value; buildChannels() end
						},
						partyLeader			= {
							type		= "toggle",
							order		= 7,
							name		= "Party Leader",
							desc		= "Party Leader Channel",
							get			= function() return db.partyLeader end,
							set			= function(_, value) db.partyLeader = value; buildChannels() end
						},
						partyLeaderText		= {
							type		= "input",
							order		= 8,
							name		= "Party Leader Text",
							desc		= "Text for party leader chat.",
							disabled	= function() return not db.channels or not db.partyLeader end,
							get			= function() return db.partyLeaderText end,
							set			= function(_, value) db.partyLeaderText = value; buildChannels() end
						},
						dungeonGuide		= {
							type		= "toggle",
							order		= 9,
							name		= "Dungeon Guide",
							desc		= "Dungeon Guide Channel",
							get			= function() return db.dungeonGuide end,
							set			= function(_, value) db.dungeonGuide = value; buildChannels() end
						},
						dungeonGuideText	= {
							type		= "input",
							order		= 10,
							name		= "Dungeon Guide Text",
							desc		= "Text for dungeon guide text.",
							disabled	= function() return not db.channels or not db.dungeonGuide end,
							get			= function() return db.dungeonGuideText end,
							set			= function(_, value) db.dungeonGuideText = value; buildChannels() end
						},
						raid				= {
							type		= "toggle",
							order		= 11,
							name		= "Raid",
							desc		= "Raid Channel",
							get			= function() return db.raid end,
							set			= function(_, value) db.raid = value; buildChannels() end
						},
						raidText			= {
							type		= "input",
							order		= 12,
							name		= "Raid Text",
							desc		= "Text for raid chat.",
							disabled	= function() return not db.channels or not db.raid end,
							get			= function() return db.raidText end,
							set			= function(_, value) db.raidText = value; buildChannels() end
						},
						raidLeader			= {
							type		= "toggle",
							order		= 13,
							name		= "Raid Leader",
							desc		= "Raid Leader Channel",
							get			= function() return db.raidLeader end,
							set			= function(_, value) db.raidLeader = value; buildChannels() end
						},
						raidLeaderText		= {
							type		= "input",
							order		= 14,
							name		= "Raid Leader Text",
							desc		= "Text for raid leader chat.",
							disabled	= function() return not db.channels or not db.raidLeader end,
							get			= function() return db.raidLeaderText end,
							set			= function(_, value) db.raidLeaderText = value; buildChannels() end
						},
						raidWarning			= {
							type		= "toggle",
							order		= 15,
							name		= "Raid Warning",
							desc		= "Raid Warning Channel",
							get			= function() return db.raidWarning end,
							set			= function(_, value) db.raidWarning = value; buildChannels() end
						},
						raidWarningText		= {
							type		= "input",
							order		= 16,
							name		= "Raid Warning Text",
							desc		= "Text for raid warning chat.",
							disabled	= function() return not db.channels or not db.raidWarning end,
							get			= function() return db.raidWarningText end,
							set			= function(_, value) db.raidWarningText = value; buildChannels() end
						},
						battleground		= {
							type		= "toggle",
							order		= 17,
							name		= "Battleground",
							desc		= "Battleground Channel",
							get			= function() return db.battleground end,
							set			= function(_, value) db.battleground = value; buildChannels() end
						},
						battlegroundText	= {
							type		= "input",
							order		= 18,
							name		= "Battleground Text",
							desc		= "Text for battleground chat.",
							disabled	= function() return not db.channels or not db.battleground end,
							get			= function() return db.battlegroundText end,
							set			= function(_, value) db.battlegroundText = value; buildChannels() end
						},
						battlegroundLeader	= {
							type		= "toggle",
							order		= 19,
							name		= "Battleground Leader",
							desc		= "Battleground Leader Channel",
							get			= function() return db.battlegroundLeader end,
							set			= function(_, value) db.battlegroundLeader = value; buildChannels() end
						},
						bgLeaderText		= {
							type		= "input",
							order		= 20,
							name		= "Battleground Leader Text",
							desc		= "Text for battleground leader chat.",
							disabled	= function() return not db.channels or not db.battlegroundLeader end,
							get			= function() return db.bgLeaderText end,
							set			= function(_, value) db.bgLeaderText = value; buildChannels() end
						},
					}
				},
				channelSoundOptions = {
					type		= "group",
					name		= "Channel Sound Options",
					disabled	= function() return not db.channelSounds end,
					args		= {
						resetSounds = {
							type	= "execute",
							order	= 100,
							name	= "Reset Sound Settings",
							desc	= "Will reset all of the sounds back to \"None\".",
							width	= "double",
							func	= function()
								for name, value in pairs(soundChannels) do
									local settingName = getSettingName(value)
									db[settingName] = "None"
								end
							end,
						}
					},
				},
				hyperlinkOptions = {
					type		= "group",
					----order		= 101,
					name		= "Hyperlink Options",
					--guiInline	= true,
					disabled	= function() return not db.hyperlinks end,
					args		= {
						combat			= {
							type	= "toggle",
							order	= 1,
							name	= "Disable in Combat",
							desc	= "Disable the hovering while in combat.",
							width	= "full",
							get		= function() return db.combat end,
							set		= function(_, value) db.combat = value end,
						},
						item			= {
							type	= "toggle",
							--order	= 2,
							name	= "Items",
							desc	= "Item hyperlinks.",
							get		= function() return db.item end,
							set		= function(_, value) db.item = value; buildHyperlinks() end,
						},
						spell			= {
							type	= "toggle",
							--order	= 3,
							name	= "Spells",
							desc	= "Spell hyperlinks.",
							get		= function() return db.spell end,
							set		= function(_, vlaue) db.spell = value; buildHyperlinks() end,
						},
						unit			= {
							type	= "toggle",
							--order	= 4,
							name	= "Units",
							desc	= "Unit hyperlinks.",
							get		= function() return db.unit end,
							set		= function(_, vlaue) db.unit = value; buildHyperlinks() end,
						},
						quest			= {
							type	= "toggle",
							--order	= 5,
							name	= "Quests",
							desc	= "Quest hyperlinks.",
							get		= function() return db.quest end,
							set		= function(_, vlaue) db.quest = value; buildHyperlinks() end,
						},
						enchant			= {
							type	= "toggle",
							--order	= 6,
							name	= "Enchants",
							desc	= "Enchant hyperlinks.",
							get		= function() return db.enchant end,
							set		= function(_, vlaue) db.enchant = value; buildHyperlinks() end,
						},
						achievement		= {
							type	= "toggle",
							--order	= 7,
							name	= "Achievements",
							desc	= "Achievement hyperlinks.",
							get		= function() return db.achievement end,
							set		= function(_, vlaue) db.achievement = value; buildHyperlinks() end,
						},
						instancelock	= {
							type	= "toggle",
							--order	= 8,
							name	= "Instance Locks",
							desc	= "Instance lock hyperlinks.",
							get		= function() return db.instancelock end,
							set		= function(_, vlaue) db.instancelock = value; buildHyperlinks() end,
						},
						talent			= {
							type	= "toggle",
							--order	= 9,
							name	= "Talents",
							desc	= "Talent hyperlinks.",
							get		= function() return db.talent end,
							set		= function(_, vlaue) db.talent = value; buildHyperlinks() end,
						},
						glyph			= {
							type	= "toggle",
							--order	= 10,
							name	= "Glyphs",
							desc	= "Glyph hyperlinks.",
							get		= function() return db.glyph end,
							set		= function(_, vlaue) db.glyph = value; buildHyperlinks() end,
						},						
					}
				},
				justifyOptions = {
					type		= "group",
					----order		= 102,
					name		= "Justification Options",
					disabled	= function() return not db.justify end,
					args		= {}
				},
				nameOptions = {
					type		= "group",
					----order		= 103,
					name		= "Name Display Options",
					childGroups	= "tab",
					disabled	= function() return not db.names end,
					args		= {
						generalOptions = {
							type		= "group",
							--order		= 1,
							name		= "General Options",
							args		= {
								nameOpen = {
									type		= "input",
									--order		= 1,
									name		= "Opening Bracket",
									desc		= "Character to use for the opening bracket.",
									get			= function() return db.nameOpen end,
									set			= function(_, value) db.nameOpen = value end,
								},
								nameClose = {
									type		= "input",
									--order		= 2,
									name		= "Closing Bracket",
									desc		= "Character to use for the closing bracket.",
									get			= function() return db.nameClose end,
									set			= function(_, value) db.nameClose = value end,
								},
								nameColorSelf = {
									type		= "toggle",
									--order		= 3,
									name		= "Color Self",
									desc		= "Color own character name in messages.",
									get			= function() return db.nameColorSelf end,
									set			= function(_, value) db.nameColorSelf = value end,
								},
								nameSelfColor = {
									type		= "color",
									--order		= 4,
									name		= "Color",
									width		= "half",
									desc		= "Color for your own character name in messages.",
									disabled	= function() return not db.nameColorSelf end,
									get			= function()
										local t = db.nameSelfColor
										return t.r, t.g, t.b
									end,
									set			= function(_, r, g, b) db.nameSelfColor = {r = r, g = g, b = b} end,
								},
								nameColorSelfReset = {
									type		= "execute",
									--order		= 5,
									name		= "Reset",
									width		= "half",
									desc		= "Reset color to default (red).",
									func		= function() db.nameSelfColor = {r = 1.0, g = 0.0, b = 0.0} end,
									confirm		= function() return "Are you sure you want to reset this color to it's default?" end,
								},
								nameEmphasizeSelf = {
									type		= "toggle",
									--order		= 6,
									name		= "Emphasize Self",
									desc		= "Add brackets to own character name in messages.",
									get			= function() return db.nameEmphasizeSelf end,
									set			= function(_, value) db.nameEmphasizeSelf = value end,
								},
								nameEmphasisColor = {
									type		= "color",
									--order		= 7,
									name		= "Color",
									width		= "half",
									desc		= "Color of the brackets used to emphasize your name.",
									disabled	= function() return not db.nameEmphasizeSelf end,
									get			= function()
										local t = db.nameEmphasisColor
										return t.r, t.g, t.b
									end,
									set			= function(_, r, g, b) db.nameEmphasisColor = {r = r, g = g, b = b} end,
								},
								nameEmphasisColorReset = {
									type		= "execute",
									--order		= 8,
									name		= "Reset",
									width		= "half",
									desc		= "Reset color to default (yellow).",
									func		= function() db.nameEmphasisColor = {r = 1.0, g = 1.0, b = 0.0} end,
									confirm		= function() return "Are you sure you want to reset this color to it's default?" end,
								},
								nameColorBy = {
									type		= "select",
									--order		= 9,
									name		= "Color Names By",
									desc		= "Choose the method to color the names by.",
									values		= {
										CLASS	= "Class",
										NAME	= "Name",
										NONE	= "None",
									},
									get			= function() return db.nameColorBy end,
									set			= function(_, value) db.nameColorBy = value end,
								},
								nameDefaultColor = {
									type		= "color",
									--order		= 10,
									name		= "Default",
									width		= "half",
									desc		= "Default color to use if all else fails.",
									disabled	= function() return db.nameColorBy == "NONE" and true or false end,
									get			= function()
										local t = db.nameDefaultColor
										return t.r, t.g, t.b
									end,
									set			= function(_, r, g, b) db.nameDefaultColor = {r = r, g = g, b = b} end
								},
								nameDefaultColorReset = {
									type		= "execute",
									order		= 11,
									name		= "Reset",
									width		= "half",
									desc		= "Reset default color to its default (grayish).",
									func		= function() db.nameDefaultColor = {r = 0.627, g = 0.627, b = 0.627} end,
									confirm		= function() return "Are you sure you want to reset the color to it's default?" end,
								},
							}
						},
						levelOptions = {
							type		= "group",
							--order		= 2,
							name		= "Level Options",
							args		= {
								level = {
									type		= "toggle",
									order		= 1,
									name		= "Include Level",
									desc		= "Include level with the name.",
									get			= function() return db.level end,
									set			= function(_, value) db.level = value end,
								},
								levelSeparator = {
									type		= "input",
									--order		= 2,
									name		= "Separator",
									desc		= "Character to use between the name and level.",
									disabled	= function() return not db.level end,
									get			= function() return db.levelSeparator end,
									set			= function(_, value) db.levelSeparator = value end,
								},
								levelMax		= {
									type		= "toggle",
									--order		= 3,
									name		= "Exclude Max Level",
									desc		= "Exclude level display for max level characters.",
									disabled	= function() return not db.level end,
									get			= function() return db.levelMax end,
									set			= function(_, value) db.levelMax = value end,
								},
								levelColor		= {
									type		= "toggle",
									--order		= 4,
									name		= "Color Levels",
									desc		= "Color levels according to quest difficulty.",
									disabled	= function() return not db.level end,
									get			= function() return db.levelColor end,
									set			= function(_, value) db.levelColor = value end,
								},
								levelLocation	= {
									type		= "select",
									name		= "Level Location",
									desc		= "Sets which side of the name to attach the level, using the separator.",
									values		= {["LEFT"] = "Left", ["RIGHT"] = "Right"},
									disabled	= function() return not db.level end,
									get			= function() return db.levelLocation end,
									set			= function(_, value) db.levelLocation = value end,
								},
							}
						},
						savingOptions = {
							type		= "group",
							--order		= 3,
							name		= "Saving Options",
							args		= {
								saveGuild = {
									type		= "toggle",
									order		= 1,
									name		= "Guild",
									desc		= "Save class data from guild.",
									get			= function() return db.saveGuild end,
									set			= function(_, value) db.saveGuild = value end,
								},
								saveGroup = {
									type		= "toggle",
									order		= 1,
									name		= "Group",
									desc		= "Save class data from groups.",
									get			= function() return db.saveGroup end,
									set			= function(_, value) db.saveGroup = value end,
								},
								saveFriend = {
									type		= "toggle",
									order		= 1,
									name		= "Friends",
									desc		= "Save class data from friends.",
									get			= function() return db.saveFriend end,
									set			= function(_, value) db.saveFriend = value end,
								},
								saveTarget = {
									type		= "toggle",
									order		= 1,
									name		= "Target/Mouseover",
									desc		= "Save class data from target/mouseover.",
									get			= function() return db.saveTarget end,
									set			= function(_, value) db.saveTarget = value end,
								},
								saveWho = {
									type		= "toggle",
									order		= 99,
									name		= "Who",
									desc		= "Save class data from /who queries.",
									get			= function() return db.saveWho end,
									set			= function(_, value) db.saveWho = value end,
								},
								saveAllWho = {
									type		= "toggle",
									order		= 99,
									name		= "All Who Data",
									desc		= "Save all data from large /who queries.",
									get			= function() return db.saveAllWho end,
									set			= function(_, value) db.saveAllWho = value end,
								},
								resetDB = {
									type		= "execute",
									order		= 100,
									width		= "double",
									name		= "Reset Data",
									desc		= "Destroy all data in the class/level table.",
									func		= function() wipe(ElvUI_ChatTweaks.db.realm.names) end,
									confirm		= function() return "Are you sure you want to clear all name/level data?" end,
								},
							}
						}
					}
				},
				timestampOptions = {
					type		= "group",
					----order		= 104,
					name		= "Timestamp Options",
					--guiInline	= true,
					disabled	= function() return not db.timestamps end,
					args		= {
						tsFormat = {
							type	= "select",
							--order	= 1,
							name	= "Timestamp Format",
							desc	= "Timestamp format.",
							values	= {
								["%I:%M:%S %p"]	= "HH:MM:SS AM (12-hour)",
								["%I:%M:S"]		= "HH:MM (12-hour)",
								["%X"]			= "HH:MM:SS (24-hour)",
								["%I:%M"]		= "HH:MM (12-hour)",
								["%H:%M"]		= "HH:MM (24-hour)",
								["%M:%S"]		= "MM:SS"
							},
							get		= function() return db.tsFormat end,
							set		= function(_, value) db.tsFormat = value end,
						},
						tsOpen	= {
							type	= "input",
							--order	= 2,
							name	= "Open",
							desc	= "Opening bracket for timestamp.  Leave blank if you don't want a bracket.",
							width	= "half",
							get		= function() return db.tsOpen end,
							set		= function(_, value) db.tsOpen = value end,
						},
						tsClose	= {
							type	= "input",
							--order	= 3,
							name	= "Close",
							desc	= "Closing bracket for timestamp.  Leave blank if you don't want a bracket.",
							width	= "half",
							get		= function() return db.tsClose end,
							set		= function(_, value) db.tsClose = value end,
						},
						tsColor = {
							type	= "color",
							--order	= 4,
							name	= "Timestamp Color",
							desc	= "The color of the timestamp.",
							get		= function()
								local t = db.tsColor
								return t.r, t.g, t.b
							end,
							set		= function(_, r, g, b) db.tsColor = {r = r, g = g, b = b} end,
						},
					}
				},
				spamFilterOptions = {
					type		= "group",
					name		= "Filtering Options",
					disabled	= function() return not db.spamFilter end,
					args		= {
						spamGoldSelling 	= {
							type		= "toggle",
							order		= 1,
							name		= "Gold Selling",
							desc		= "Filters gold selling, powerleveling, and other services that are against Blizzard's EULA.",
							get			= function() return db.spamGoldSelling end,
							set			= function(_, value) db.spamGoldSelling = value end,
						},
						spamGuildRecruit	= {
							type		= "toggle",
							order		= 2,
							name		= "Guild Recruitment",
							desc		= "Filters guild recruitment messages.\n\n|cffff0000DOES NOT REPORT THEM!|r",
							get			= function() return db.spamGuildRecruit end,
							set			= function(_, value) db.spamGuildRecruit = value end,
						},
						spamRemoveIcons	= {
							type		= "toggle",
							order		= 3,
							name		= "Remove Icons",
							desc		= "Removes icons from messages to prevent the various objects people try to make.",
							get			= function() return db.spamRemoveIcons end,
							set			= function(_, value) db.spamRemoveIcons = value end,
						},
						spamReportingOptions	= {
							type		= "group",
							guiInline	= true,
							order		= 100,
							name		= "Reporting Options",
							disabled	= function() return not db.spamGoldSelling end,
							args		= {
								spamAutoReport		= {
									type		= "toggle",
									order		= 1,
									name		= "Report",
									desc		= "Auto report anyone reaching the spam threshold (3 points).",
									get			= function() return db.spamAutoReport end,
									set			= function(_, value) db.spamAutoReport = value end,
								},
								spamReportConfirm	= {
									type		= "toggle",
									order		= 2,
									name		= "Confirm Report?",
									desc		= "Confirm reporting before actually reporting them.",
									disabled	= function() return not db.spamAutoReport end,
									get			= function() return db.spamReportConfirm end,
									set			= function(_, value) db.spamReportConfirm = value end,
								},
								spamReportSilent	= {
									type		= "toggle",
									order		= 3,
									name		= "Silent Report",
									desc		= "Surpress the reporting system message.",
									disabled	= function() return not db.spamAutoReport end,
									get			= function() return db.spamReportSilent end,
									set			= function(_, value) db.spamReportSilent = value end,
								},
							}
						},
						spamKeywordOptions	= {
							type		= "group",
							guiInline	= true,
							order		= 101,
							name		= "Keyword Filtering",
							args		= {
								spamKeywords	= {
									type		= "toggle",
									order		= 1,
									width		= "double",
									name		= "Keywords or Phrases",
									desc		= "Filter certain words or phrases according to your settings.\n\n|cffff0000DOES NOT REPORT THEM!|r",
									get			= function() return db.spamKeywords end,
									set			= function(_, value) db.spamKeywords = value end,
								},
								spamFilters		= {
									type		= "input",
									order		= 2,
									multiline	= true,
									width		= "double",
									name		= "Triggers",
									desc		= "Words or phrases that will trigger the filter.",
									disabled	= function() return not db.spamKeywords end,
									get			= function() return ElvUI_ChatTweaks:FiltersToString() end,
									set			= function(_, value)
										local trigList = {strsplit("\n", value:trim())}
										ElvUI_ChatTweaks:PopulateFilters(trigList)
									end,
								},
							}
						},
						spamMiscOptions = {
							type		= "group",
							guiInline	= true,
							order		= 102,
							name		= "Miscellaneous Filtering",
							args		= {
								spamAFK	= {
									type	= "toggle",
									name	= "Filter AFK Messages",
									desc	= "Filter automatic AFK messages.",
									get		= function() return db.spamAFK end,
									set		= function(_, value) db.spamAFK = value end,
								},
								spamDND	= {
									type	= "toggle",
									name	= "Filter DND Messages",
									desc	= "Filter automatic DND messages.",
									get		= function() return db.spamDND end,
									set		= function(_, value) db.spamDND = value end,
								},
							},
						},
					}
				},
				rollFrameOptions = {
					type		= "group",
					name		= "Roll Frame Options",
					disabled	= function() return not db.rollFrame end,
					args		= {
						rollRouteTo = {
							type		= "select",
							order		= 1,
							name		= "Route To",
							desc		= "Route the rolls to this chat frame.  Setting this will automatically send all other system messages to ChatFrame1 (main chat frame).",
							values		= ElvUI_ChatTweaks:GetPossibleRollFrames(),
							get			= function() return db.rollRouteTo end,
							set			= function(_, value) db.rollRouteTo = value end
						},
						rollFlash = {
							type		= "toggle",
							order		= 2,
							name		= "Flash Roll Frame",
							desc		= "Flashes the roll frame when a new roll is received.",
							get			= function() return db.rollFlash end,
							set			= function(_, value) db.rollFlash = value end,
						},
						rollColors = {
							type		= "toggle",
							order		= 3,
							name		= "Custom Roll Colors",
							desc		= "Allows you to customize the colors of the rolls.",
							get			= function() return db.rollColors end,
							set			= function(_, value) db.rollColors = value end,
						},
						rollColorOptions = {
							type		= "group",
							order		= 100,
							guiInline	= true,
							disabled	= function() return not db.rollColors end,
							name		= "Roll Colors",
							args		= {
								rollColorReg = {
									type		= "toggle",
									order		= 1,
									name		= "Color Rolls",
									desc		= "Color rolls different than the default yellow.",
									get			= function() return db.rollColorReg end,
									set			= function(_, value) db.rollColorReg = value end,
								},
								rollRegColor = {
									type		= "color",
									order		= 2,
									name		= "Regular Roll Color",
									desc		= "Color to use for regular rolls.",
									disabled	= function() return not db.rollColors or not db.rollColorReg end,
									get			= function() return db.rollRegColor.r, db.rollRegColor.b, db.rollRegColor.b end,
									set			= function(_, r, g, b) db.rollRegColor = {r = r, g = g, b = b} end,
								},
								rollColorLoss = {
									type		= "toggle",
									order		= 3,
									name		= "Color Losses",
									desc		= "Color someone else's loss of the roll game.",
									get			= function() return db.rollColorLoss end,
									set			= function(_, value) db.rollColorLoss = value end,
								},
								rollLossColor = {
									type		= "color",
									order		= 4,
									name		= "Loss Color",
									desc		= "Color to use when someone else loses the roll game.",
									disabled	= function() return not db.rollColors or not db.rollColorLoss end,
									get			= function() return db.rollLossColor.r, db.rollLossColor.g, db.rollLossColor.b end,
									set			= function(_, r, g, b) db.rollLossColor = {r = r, g = g, b = b} end,
								},
								rollColorOwn = {
									type		= "toggle",
									order		= 5,
									name		= "Color Own Rolls",
									desc		= "Colors your own rolls to make them stand out.",
									get			= function() return db.rollColorOwn end,
									set			= function(_, value) db.rollColorOwn = value end,
								},
								rollOwnColor = {
									type		= "color",
									order		= 6,
									name		= "Own Rolls Color",
									desc		= "The color of your rolls.",
									disabled	= function() return not db.rollColors or not db.rollColorOwn end,
									get			= function() return db.rollOwnColor.r, db.rollOwnColor.g, db.rollOwnColor.b end,
									set			= function(_, r, g, b) db.rollOwnColor = {r = r, g = g, b = b} end,
								},
								rollColorOwnLoss = {
									type		= "toggle",
									order		= 7,
									name		= "Color Own Loss",
									desc		= "Color your own loss of the roll game.",
									get			= function() return db.rollColorOwnLoss end,
									set			= function(_, value) db.rollColorOwnLoss = value end,
								},
								rollOwnLossColor = {
									type		= "color",
									order		= 8,
									name		= "Own Loss Color",
									desc		= "Color to use when you lose the roll game.",
									disabled	= function() return not db.rollColors or not db.rollColorOwnLoss end,
									get			= function() return db.rollOwnLossColor.r, db.rollOwnLossColor.g, db.rollOwnLossColor.b end,
									set			= function(_, r, g, b) db.rollOwnLossColor = {r = r, g = g, b = b} end,
								},
							}
						},
					}
				},
			}
		}
	}
}